#!/usr/bin/env python3
#
# Sniffles2
# A fast structural variant caller for long-read sequencing data
#
# Created: 29.08.2021
# Author:  Moritz Smolka
# Contact: moritz.g.smolka@gmail.com
#

from dataclasses import dataclass
import statistics
import math

from sniffles import sv
from sniffles import leadprov

@dataclass
class Cluster:
    id: str
    svtype: str
    contig: str
    start: int
    end: int
    seed: int
    leads: list
    repeat: bool
    leads_long: list

    def compute_metrics(self,max_n=100):
        self.span=self.end-self.start

        n=min(len(self.leads),max_n)
        if n==0:
            self.mean_svlen=0
            self.stdev_start=0
            return

        step=int(len(self.leads)/n)
        if n>1:
            self.mean_svlen=sum(self.leads[i].svlen for i in range(0,len(self.leads),step))/float(n)
            self.stdev_start=statistics.stdev(self.leads[i].ref_start for i in range(0,len(self.leads),step))
        else:
            self.mean_svlen=self.leads[0].svlen
            self.stdev_start=0

def merge_inner(cluster,threshold):
    read_seq={}
    for ld in cluster.leads:
        if not ld.read_qname in read_seq:
            read_seq[ld.read_qname]=[]
        read_seq[ld.read_qname].append(ld)

    cluster.leads=[]
    for qname in read_seq:
        read_seq[qname].sort(key=lambda k: k.ref_start)
        to_merge=read_seq[qname][0]

        curr_lead=to_merge

        last_ref_end=to_merge.ref_end
        last_qry_end=to_merge.qry_end
        last_ref_start=to_merge.ref_start
        last_qry_start=to_merge.qry_start

        for to_merge in read_seq[qname][1:]:
            merge=(threshold==-1) or ((abs(to_merge.ref_start-last_ref_end) < threshold or abs(to_merge.ref_start-last_ref_start) < threshold) and (abs(to_merge.qry_start-last_qry_end) < threshold or abs(to_merge.qry_start-last_qry_start) < threshold))
            if merge:
                curr_lead.svlen+=to_merge.svlen
                if to_merge.seq==None or curr_lead.seq==None:
                    curr_lead.seq=None
                else:
                    curr_lead.seq+=to_merge.seq
            else:
                cluster.leads.append(curr_lead)
                curr_lead=to_merge
            last_ref_end=to_merge.ref_end
            last_qry_end=to_merge.qry_end
            last_ref_start=to_merge.ref_start
            last_qry_start=to_merge.qry_start

        cluster.leads.append(curr_lead)
    return cluster

def resplit(cluster,prop,binsize,merge_threshold_min,merge_threshold_frac):
    bins_leads={}
    for lead in cluster.leads:
        bin=int(abs(prop(lead))/binsize)*binsize
        if not bin in bins_leads:
            bins_leads[bin]=[lead]
        else:
            bins_leads[bin].append(lead)

    new_clusters=list(sorted(bins_leads.keys()))
    i=1
    while len(new_clusters) > 1 and i < len(new_clusters):
        last_cluster=new_clusters[i-1]
        curr_cluster=new_clusters[i]
        merge_threshold=max(merge_threshold_min,min(curr_cluster,last_cluster)*merge_threshold_frac)
        merge=abs(curr_cluster - last_cluster) <= merge_threshold
        if merge:
            bins_leads[new_clusters[i]].extend(bins_leads[new_clusters[i-1]])
            new_clusters.pop(i-1)
            i=max(0,i-2)
        else:
            i+=1

    for cluster_index in new_clusters:
        new_cluster=Cluster(id=cluster.id+f".{cluster_index}",
                            svtype=cluster.svtype,
                            contig=cluster.contig,
                            start=cluster.start,
                            end=cluster.end,
                            seed=cluster.seed,
                            leads=bins_leads[cluster_index],
                            repeat=cluster.repeat,
                            leads_long=cluster.leads_long)
        yield new_cluster

def resplit_bnd(cluster,merge_threshold):
    contigs_leads={}

    #Split by mate contig and mate ref start
    for lead in cluster.leads:
        if not lead.bnd_info.mate_contig in contigs_leads:
            contigs_leads[lead.bnd_info.mate_contig]={}

        bin=int(lead.bnd_info.mate_ref_start/merge_threshold)*merge_threshold if merge_threshold>0 else 0
        if not bin in contigs_leads[lead.bnd_info.mate_contig]:
            contigs_leads[lead.bnd_info.mate_contig][bin]=[lead]
        else:
            contigs_leads[lead.bnd_info.mate_contig][bin].append(lead)

    for contig in contigs_leads:
        bins=sorted(contigs_leads[contig])
        curr_leads=[]+contigs_leads[contig][bins[0]]
        last_bin=bins[0]
        for bin in bins[1:]:
            if bin - last_bin <= merge_threshold:
                curr_leads.extend(contigs_leads[contig][bin])
            else:
                if len(curr_leads):
                    new_cluster=Cluster(id=cluster.id+f".CHR2.{contig}.POS2.{bin}",
                                        svtype=cluster.svtype,
                                        contig=cluster.contig,
                                        start=cluster.start,
                                        end=cluster.end,
                                        seed=cluster.seed,
                                        leads=[k for k in curr_leads],
                                        repeat=cluster.repeat,
                                        leads_long=None)
                    yield new_cluster
                curr_leads=[]+contigs_leads[contig][bin]
            last_bin=bin
        if len(curr_leads):
            new_cluster=Cluster(id=cluster.id+f".CHR2.{contig}.POS2.{bin}",
                                svtype=cluster.svtype,
                                contig=cluster.contig,
                                start=cluster.start,
                                end=cluster.end,
                                seed=cluster.seed,
                                leads=[k for k in curr_leads],
                                repeat=cluster.repeat,
                                leads_long=None)
            curr_leads=[]
            yield new_cluster

def resolve(svtype,leadtab_provider,config,tr):
    leadtab=leadtab_provider.leadtab[svtype]
    seeds=sorted(leadtab_provider.leadtab[svtype])

    if len(seeds)==0:
        return []

    #Initialize tandem repeat region handling
    if tr != None:
        tr_index=0
        if len(tr)==0:
            tr=None
        else:
            tr_start,tr_end=tr[tr_index]

    clusters=[]
    for seed_index,seed in enumerate(seeds):
        if config.dev_call_region != None:
            if seed<config.dev_call_region["start"] or seed>config.dev_call_region["end"]:
                continue

        within_tr=False
        if tr!=None and tr_index < len(tr):
            while tr_end < seed and tr_index+1 < len(tr):
                tr_index+=1
                tr_start,tr_end=tr[tr_index]
            if seed > tr_start and seed < tr_end:
                within_tr=True

        if svtype=="INS":
            leads=[lead for lead in leadtab[seed] if lead.svlen!=None]
            leads_long=[lead for lead in leadtab[seed] if lead.svlen==None]
        else:
            leads=leadtab[seed]
            leads_long=None

        cluster=Cluster(id=f"CL.{svtype}.{leadtab_provider.contig}.{leadtab_provider.start}.{seed_index}",
                        svtype=svtype,
                        contig=leadtab_provider.contig,
                        start=seed,
                        end=seed+config.cluster_binsize,
                        seed=seed,
                        leads=leads,
                        repeat=within_tr or config.repeat,
                        leads_long=leads_long)

        cluster.compute_metrics()
        clusters.append(cluster)


    #Merge clusters
    cluster_count_initial=len(clusters)
    i=0
    while i < len(clusters) - 1:
        curr_cluster=clusters[i]
        next_cluster=clusters[i+1]

        #assert((next_cluster.end - curr_cluster.start) >= 0)
        #assert((next_cluster.start - curr_cluster.end) >= 0)

        #Fast Adaptive Scanning Clustering
        #  Inter-Cluster distance < (StdDev-Start-Within-A + StdDev-Stat-Within-B) * r   (r=1.0 default)
        #  Merged size (OR: Inter-cluster distance) < (MeanSVlen between both clusters) * h (h=0.5 default)
        #  Max. overall cluster size criterion?

        inner_dist=(next_cluster.start - curr_cluster.end)
        outer_dist=(next_cluster.end - curr_cluster.start)
        merge = inner_dist <= min(curr_cluster.stdev_start, next_cluster.stdev_start) * config.cluster_r
        merge = merge or ( (config.repeat or curr_cluster.repeat or next_cluster.repeat) and outer_dist <= min(config.cluster_repeat_h_max, (abs(curr_cluster.mean_svlen)+abs(next_cluster.mean_svlen)) * config.cluster_repeat_h) )
        merge = merge or (svtype=="BND" and inner_dist <= config.cluster_merge_bnd)

        if merge:
            clusters.pop(i+1)
            curr_cluster.leads+=next_cluster.leads
            if svtype=="INS":
                curr_cluster.leads_long+=next_cluster.leads_long
            curr_cluster.end=next_cluster.end
            curr_cluster.repeat=curr_cluster.repeat or next_cluster.repeat
            curr_cluster.compute_metrics()
            i=max(0,i-2)
        i+=1

    if config.dev_dump_clusters:
        filename=f"{config.input}.clusters.{svtype}.{leadtab_provider.contig}.{leadtab_provider.start}.{leadtab_provider.end}.bed"
        print(f"Dumping clusters to {filename}")
        with open(filename,"w") as h:
            for c in clusters:
                info=f"ID={c.id}, #LEADS={len(c.leads)}; "
                for ld in c.leads:
                    info+=f"(ref_start={ld.ref_start},svlen={ld.svlen},source={ld.source}); "
                h.write(f"{c.contig}\t{c.start}\t{c.end}\t\"{info}\"\n")

    for cluster in clusters:
        if len(cluster.leads)==0:
            continue

        if svtype == "BND":
            if config.dev_no_resplit:
                yield cluster
            else:
                for new_cluster in resplit_bnd(cluster,merge_threshold=config.bnd_cluster_resplit):
                    yield new_cluster
        else:
            if svtype=="INS" or svtype=="DEL":
                if cluster.repeat:
                    merge_inner_threshold=-1
                else:
                    merge_inner_threshold=config.cluster_merge_pos

                merge_inner(cluster,merge_inner_threshold)

            if not config.dev_no_resplit_repeat and not config.dev_no_resplit:
                for new_cluster in resplit(cluster,
                                           prop=lambda lead: lead.svlen,
                                           binsize=config.cluster_resplit_binsize,
                                           merge_threshold_min=config.minsvlen,
                                           merge_threshold_frac=config.cluster_merge_len):
                    yield new_cluster
            else:
                yield cluster

def resolve_block_groups(svtype,svcands,groups_initial,config):
    """For clustering groups of SVs for combining .snfs (multi-call)"""

    #TODO: Remove sorting
    groups=groups_initial
    for svcand in sorted(svcands,key=lambda cand: cand.support,reverse=True):
        best_group=None
        best_dist=math.inf

        if svtype=="BND":
            mate_contig,mate_ref_start=svcand.bnd_info.mate_contig,svcand.bnd_info.mate_ref_start
            for group in groups:
                #TODO: Favor bigger groups in placement
                dist=abs(group.pos_mean - svcand.pos) + abs(group.bnd_mate_ref_start_mean - mate_ref_start)
                if dist < best_dist and dist <= config.cluster_merge_bnd*2 and group.bnd_mate_contig==mate_contig:
                    if not config.combine_separate_intra or not svcand.sample_internal_id in group.included_samples:
                        best_group=group
                        best_dist=dist
        else:
            for group in groups:
                #TODO: Favor bigger groups in placement
                dist=abs(group.pos_mean - svcand.pos) + abs(abs(group.len_mean) - abs(svcand.svlen))
                minlen=float(min(abs(group.len_mean),abs(svcand.svlen)))
                if minlen>0 and dist < best_dist and dist <= config.combine_match * math.sqrt(minlen) and dist <= config.combine_match_max:
                    if not config.combine_separate_intra or not svcand.sample_internal_id in group.included_samples:
                        best_group=group
                        best_dist=dist

        if best_group==None:
            group=sv.SVGroup(candidates=[svcand],
                             pos_mean=float(svcand.pos),
                             len_mean=float(abs(svcand.svlen)),
                             included_samples=set([svcand.sample_internal_id]),
                             coverages_nonincluded=dict())
            if svtype=="BND":
                group.bnd_mate_contig=mate_contig
                group.bnd_mate_ref_start_mean=mate_ref_start
            groups.append(group)
        else:
            group_size=len(best_group.candidates)
            best_group.pos_mean*=group_size
            best_group.len_mean*=group_size
            best_group.pos_mean+=svcand.pos
            best_group.len_mean+=abs(svcand.svlen)
            if svtype=="BND":
                best_group.bnd_mate_ref_start_mean*=group_size
                best_group.bnd_mate_ref_start_mean+=mate_ref_start

            best_group.candidates.append(svcand)
            group_size+=1
            best_group.pos_mean/=group_size
            best_group.len_mean/=group_size
            best_group.included_samples.add(svcand.sample_internal_id)

            if svtype=="BND":
                best_group.bnd_mate_ref_start_mean/=group_size

    return groups



#!/usr/bin/env python3
#
# Sniffles2
# A fast structural variant caller for long-read sequencing data
#
# Created: 15.08.2021
# Author:  Moritz Smolka
# Contact: moritz.g.smolka@gmail.com
#

import pysam
import os

from sniffles import sv
from sniffles import util

def format_info(k,v):
    if isinstance(v,float):
        return f"{k}={v:.3f}"
    elif isinstance(v,list):
        return f"{k}={','.join(v)}"
    else:
        return f"{k}={v}"

def format_genotype(gt):
    if len(gt)==6:
        a,b,qual,dr,dv,ps=gt
        if ps!=None and (a,b)==(0,1):
            gt_sep="|"
            if ps==1:
                a,b=b,a
        else:
            gt_sep="/"
        return f"{a}{gt_sep}{b}:{qual}:{dr}:{dv}"
    else:
        a,b,qual,dr,dv,ps,id=gt
        if ps!=None and (a,b)==(0,1):
            gt_sep="|"
            if ps==1:
                a,b=b,a
        else:
            gt_sep="/"
        return f"{a}{gt_sep}{b}:{qual}:{dr}:{dv}:{id}"

class VCF:
    def __init__(self,config,handle):
        self.config=config
        self.handle=handle
        self.call_count=0
        self.info_order=["SVTYPE","SVLEN","END","SUPPORT","RNAMES","COVERAGE","STRAND"]
        if config.qc_nm:
            self.info_order.append("NM")

        self.default_genotype=config.genotype_none

        if config.mode=="combine":
            self.genotype_format=config.genotype_format+":ID"
            self.default_genotype+=tuple(["NULL"])
        else:
            self.genotype_format=config.genotype_format

        self.reference_handle=None

    def open_reference(self):
        if self.config.reference == None:
            return
        if not os.path.exists(self.config.reference+".fai") and not os.path.exists(self.config.reference+".gzi"):
            print(f"Info: Fasta index for {self.config.reference} not found. Generating with pysam.faidx (this may take a while)")
            pysam.faidx(self.config.reference)
        self.reference_handle=pysam.FastaFile(self.config.reference)

    def write_header(self,contigs_lengths):
        self.write_header_line("fileformat=VCFv4.2")
        self.write_header_line(f"source={self.config.version}_{self.config.build}")
        self.write_header_line('command="'+self.config.command+'"')
        self.write_header_line('fileDate="'+self.config.start_date+'"')
        for contig,contig_len in contigs_lengths:
            self.write_header_line(f"contig=<ID={contig},length={contig_len}>")

        self.write_header_line('ALT=<ID=INS,Description="Insertion">')
        self.write_header_line('ALT=<ID=DEL,Description="Deletion">')
        self.write_header_line('ALT=<ID=DUP,Description="Duplication">')
        self.write_header_line('ALT=<ID=INV,Description="Inversion">')
        self.write_header_line('ALT=<ID=BND,Description="Breakend; Translocation">')

        self.write_header_line('FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">')
        self.write_header_line('FORMAT=<ID=GQ,Number=1,Type=Integer,Description="Genotype quality">')
        self.write_header_line('FORMAT=<ID=DR,Number=1,Type=Integer,Description="Number of reference reads">')
        self.write_header_line('FORMAT=<ID=DV,Number=1,Type=Integer,Description="Number of variant reads">')
        self.write_header_line('FORMAT=<ID=ID,Number=1,Type=String,Description="Individual sample SV ID for multi-sample output">')

        self.write_header_line('FILTER=<ID=PASS,Description="All filters passed">')
        self.write_header_line('FILTER=<ID=GT,Description="Genotype filter">')
        self.write_header_line('FILTER=<ID=SUPPORT_MIN,Description="Minimum read support filter">')
        self.write_header_line('FILTER=<ID=STDEV_POS,Description="SV Breakpoint standard deviation filter">')
        self.write_header_line('FILTER=<ID=STDEV_LEN,Description="SV length standard deviation filter">')
        self.write_header_line('FILTER=<ID=COV_MIN,Description="Minimum coverage filter">')
        self.write_header_line('FILTER=<ID=COV_CHANGE,Description="Coverage change filter">')
        self.write_header_line('FILTER=<ID=STRAND,Description="Strand support filter">')
        self.write_header_line('FILTER=<ID=SVLEN_MIN,Description="SV length filter">')
        self.write_header_line('FILTER=<ID=NM,Description="Alignment noise level filter">')
        self.write_header_line('INFO=<ID=PRECISE,Number=0,Type=Flag,Description="Structural variation with precise breakpoints">')
        self.write_header_line('INFO=<ID=IMPRECISE,Number=0,Type=Flag,Description="Structural variation with imprecise breakpoints">')
        self.write_header_line('INFO=<ID=SVLEN,Number=1,Type=Integer,Description="Length of structural variation">')
        self.write_header_line('INFO=<ID=SVTYPE,Number=1,Type=String,Description="Type of structural variation">')
        self.write_header_line('INFO=<ID=CHR2,Number=1,Type=String,Description="Mate chromsome for BND SVs">')
        self.write_header_line('INFO=<ID=SUPPORT,Number=1,Type=Integer,Description="Number of reads supporting the structural variation">')
        self.write_header_line('INFO=<ID=SUPPORT_INLINE,Number=1,Type=Integer,Description="Number of reads supporting an INS/DEL SV (non-split events only)">')
        self.write_header_line('INFO=<ID=SUPPORT_LONG,Number=1,Type=Integer,Description="Number of soft-clipped reads putatively supporting the long insertion SV">')
        self.write_header_line('INFO=<ID=END,Number=1,Type=Integer,Description="End position of structural variation">')
        self.write_header_line('INFO=<ID=STDEV_POS,Number=1,Type=Float,Description="Standard deviation of structural variation start position">')
        self.write_header_line('INFO=<ID=STDEV_LEN,Number=1,Type=Float,Description="Standard deviation of structural variation length">')
        self.write_header_line('INFO=<ID=COVERAGE,Number=.,Type=Float,Description="Coverages near upstream, start, center, end, downstream of structural variation">')
        self.write_header_line('INFO=<ID=STRAND,Number=1,Type=String,Description="Strands of supporting reads for structural variant">')
        self.write_header_line('INFO=<ID=AC,Number=.,Type=Integer,Description="Allele count, summed up over all samples">')
        self.write_header_line('INFO=<ID=SUPP_VEC,Number=1,Type=String,Description="List of read support for all samples">')
        #self.write_header_line('INFO=<ID=COVERAGE_REGIONAL,Number=A,Type=Float,Description="Coverage for local region">')
        #self.write_header_line('INFO=<ID=COVERAGE_GLOBAL,Number=A,Type=Float,Description="Coverage for global region">')
        #self.write_header_line('INFO=<ID=MINSUPPORT,Number=A,Type=Integer,Description="Determined minimum support threshold for this region">')
        #self.write_header_line('INFO=<ID=COVERAGE_UPSTREAM,Number=A,Type=Float,Description="Coverage upstream of structural variation start">')
        #self.write_header_line('INFO=<ID=COVERAGE_DOWNSTREAM,Number=A,Type=Float,Description="Coverage downstream of structural variation end">')
        self.write_header_line('INFO=<ID=CONSENSUS_SUPPORT,Number=1,Type=Integer,Description="Number of reads that support the generated insertion (INS) consensus sequence">')
        self.write_header_line('INFO=<ID=RNAMES,Number=.,Type=String,Description="Names of supporting reads (if enabled with --output-rnames)">')
        self.write_header_line('INFO=<ID=AF,Number=1,Type=Float,Description="Allele Frequency">')
        self.write_header_line('INFO=<ID=NM,Number=.,Type=Float,Description="Mean number of query alignment length adjusted mismatches of supporting reads">')
        self.write_header_line('INFO=<ID=PHASE,Number=.,Type=String,Description="Phasing information derived from supporting reads, represented as list of: HAPLOTYPE,PHASESET,HAPLOTYPE_SUPPORT,PHASESET_SUPPORT,HAPLOTYPE_FILTER,PHASESET_FILTER">')
        #self.write_header_line('INFO=<ID=DBG_CLUSTER_SPAN,Number=A,Type=Float,Description="Debug info">')

        samples_header="\t".join(sample_id for internal_id,sample_id in self.config.sample_ids_vcf)
        self.write_raw(f"#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t{samples_header}")

    def write_raw(self,text,endl="\n"):
        if self.config.vcf_output_bgz:
            self.handle.write(text.encode())
            self.handle.write(endl.encode())
        else:
            self.handle.write(text)
            self.handle.write(endl)

    def write_header_line(self,text):
        self.write_raw("##"+text)

    def write_call(self,call):
        #pysam coordinates are 0-based, VCF 1-based - therefore +1 is added to call.pos
        end=call.end + 1
        pos=call.pos + 1

        #Determine genotypes columns
        ac=0 #Allele count
        supvec=[]
        sample_genotypes=[]
        for internal_id, sample_id in self.config.sample_ids_vcf:
            if internal_id in call.genotypes and call.genotypes[internal_id]!=None:
                gt_curr=call.genotypes[internal_id]
                sample_genotypes.append(format_genotype(gt_curr))
                if gt_curr[0]!="." and gt_curr[4]>0: #Not non-genotype and has supporting reads
                    ac+=sum(call.genotypes[internal_id][:2])
                    supp="1"
                else:
                    supp="0"
            else:
                sample_genotypes.append(format_genotype(self.default_genotype))
                supp="0"
            supvec.append(supp)

        if len(self.config.sample_ids_vcf) > 1:
            call.set_info("AC",ac)
            call.set_info("SUPP_VEC","".join(supvec))
            if ac==0:
                call.filter="GT"

        #Output core SV attributes
        infos={"SVTYPE": call.svtype,
               "SVLEN": call.svlen,
               "END": end,
               "SUPPORT": call.support,
               "RNAMES": call.rnames if self.config.output_rnames else None,
               "COVERAGE": f"{call.coverage_upstream},{call.coverage_start},{call.coverage_center},{call.coverage_end},{call.coverage_downstream}",
               "STRAND": ("+" if call.fwd>0 else "") + ("-" if call.rev>0 else ""),
               "NM": call.nm}

        if call.svtype=="BND":
            infos["SVLEN"]=None
            infos["END"]=None

        infos_ordered=["PRECISE" if call.precise else "IMPRECISE"]
        infos_ordered.extend(format_info(k,infos[k]) for k in self.info_order if infos[k]!=None)
        info_str=";".join(infos_ordered)

        #Output call specific additional information
        for k in sorted(call.info):
            if call.info[k]==None:
                continue
            info_str+=";" + format_info(k,call.info[k])

        #if call.id==None:
        #    call.id=f"Sniffles2.{call.svtype}.{self.call_count+1:06}"

        #Resolve DEL sequence
        if not self.config.symbolic and call.svtype=="DEL" and self.reference_handle != None and abs(call.svlen) <= self.config.max_del_seq_len:
            try:
                call.ref=self.reference_handle.fetch(call.contig,call.pos,call.pos-call.svlen)
                call.alt="N"
            except KeyError:
                call.ref="N"
                call.alt=f"<{call.svtype}>"
            except ValueError:
                call.ref="N"
                call.alt=f"<{call.svtype}>"

        if self.config.symbolic:
            call.ref="N"
            call.alt=f"<{call.svtype}>"

        call.qual=max(0,min(60,call.qual))

        self.write_raw("\t".join(str(v) for v in [call.contig,pos,self.config.id_prefix+call.id,call.ref,call.alt,call.qual,call.filter,info_str,self.genotype_format]+sample_genotypes))
        self.call_count+=1

    def read_svs_iter(self):
        self.header_str=""
        line_index=0
        for line in self.handle:
            try:
                if isinstance(line,bytes):
                    line=line.decode("utf-8")
                line_index+=1
                line_strip=line.strip()
                if line_strip=="" or line_strip[0]=="#":
                    if line_strip[0]=="#":
                        self.header_str+=line_strip+"\n"
                    continue
                CHROM,POS,ID,REF,ALT,QUAL,FILTER,INFO=line.split("\t")[:8]
                info_dict={}
                for info_item in INFO.split(";"):
                    if "=" in info_item:
                        key,value=info_item.split("=")
                    else:
                        key,value=info_item,True
                    info_dict[key]=value
                call=sv.SVCall(contig=CHROM,
                               pos=int(POS)-1,
                               id=line_index,
                               ref=REF,
                               alt=ALT,
                               qual=QUAL,
                               filter=FILTER,
                               info=info_dict,
                               svtype=None,
                               svlen=None,
                               end=None,
                               rnames=None,
                               qc=True,
                               postprocess=None,
                               genotypes=None,
                               precise=None,
                               support=0,
                               fwd=0,
                               rev=0,
                               nm=-1)
                if len(call.alt)>len(call.ref):
                    call.svtype="INS"
                    call.svlen=len(call.alt)
                    call.end=call.pos
                else:
                    call.svtype="DEL"
                    call.svlen=-len(call.ref)
                    call.end=call.pos+call.svlen

                if "SVTYPE" in info_dict:
                    call.svtype=info_dict["SVTYPE"]
                    if call.svtype=="TRA":
                        call.svtype="BND"

                if "SVLEN" in info_dict:
                    call.svlen=int(info_dict["SVLEN"])
                if "END" in info_dict:
                    call.end=int(info_dict["END"])

                if call.svtype=="BND":
                    bnd_parts=call.alt.replace("]","[").split("[")
                    if len(bnd_parts)>2:
                        mate_contig,mate_ref_start=bnd_parts[1].split(":")
                        call.bnd_info=sv.SVCallBNDInfo(mate_contig=mate_contig, mate_ref_start=int(mate_ref_start), is_first=(call.alt[0]=="N"), is_reverse=("]" in call.alt))
                    else:
                        raise ValueError("BND ALT not formatted according to VCF 4.2 specifications")

                call.raw_vcf_line=line_strip
                call.raw_vcf_line_index=line_index
                yield call
            except Exception as e:
                util.fatal_error(f"Error parsing input VCF: Line {line_index}: {e}")

    def rewrite_genotype(self,svcall):
        parts_no_gt=svcall.raw_vcf_line.split("\t")[:8]
        gt_format=self.config.genotype_format
        if svcall.genotype_match_sv != None:
            if len(svcall.genotype_match_sv.genotypes)>0:
                gt=svcall.genotype_match_sv.genotypes[0]
            else:
                gt=svcall.genotypes[0]
        else:
            gt=svcall.genotypes[0] #0/0 or ./. depending on whether input SV had coverage in sample
        #parts=parts_no_gt + [gt_format,vcf.format_genotype(gt)]
        #gt_vcf=svcall.raw_vcf_line.split("\t")[9].split(":")[0]
        #parts= parts_no_gt + [gt_vcf] + [gt_format,vcf.format_genotype(gt)]
        parts=parts_no_gt + [gt_format,format_genotype(gt)]
        #parts[7]="NA"
        #parts[3]=f"REF_{len(parts[3])}"
        #parts[4]=f"ALT_{len(parts[4])}"
        self.write_raw("\t".join(parts))

    def rewrite_header_genotype(self,orig_header):
        header_lines=orig_header.split("\n")
        header_lines.insert(1,'##genotypeFileDate="'+self.config.start_date+'"')
        header_lines.insert(1,'##genotypeCommand="'+self.config.command+'"')
        header_lines.insert(1,f"##genotypeSource={self.config.version}_{self.config.build}")
        self.write_raw("\n".join(header_lines),endl="")

    def close(self):
        self.handle.close()



#!/usr/bin/env python3
#
# Sniffles2
# A fast structural variant caller for long-read sequencing data
#
# Created: 26.09.2021
# Author:  Moritz Smolka
# Contact: moritz.g.smolka@gmail.com
#

import pickle
import json
import gzip
import math

from sniffles import sv

class SNFile:
    def __init__(self,config,handle,filename=None):
        self.config=config
        self.handle=handle
        self.filename=filename
        self.blocks={}
        self.index={}
        self.total_length=0

    def store(self,svcand):
        block_index=int(svcand.pos/self.config.snf_block_size)*self.config.snf_block_size
        if not block_index in self.blocks:
            self.blocks[block_index]={svtype: [] for svtype in sv.TYPES}
            self.blocks[block_index]["_COVERAGE"]={}
        if not self.config.output_rnames:
            svcand.rnames=None
        self.blocks[block_index][svcand.svtype].append(svcand)

    def annotate_block_coverages(self,lead_provider,resolution=500):
        config=self.config
        start_bin=lead_provider.covrtab_min_bin
        end_bin=int(lead_provider.end/config.coverage_binsize)*config.coverage_binsize
        coverage_fwd=0
        coverage_rev=0

        coverage_sum=0
        bin_count=0

        coverage_binsize_combine=self.config.coverage_binsize_combine
        snf_block_size=config.snf_block_size

        for bin in range(start_bin, end_bin+config.coverage_binsize,config.coverage_binsize):
            if bin in lead_provider.covrtab_fwd:
                coverage_fwd+=lead_provider.covrtab_fwd[bin]

            if bin in lead_provider.covrtab_rev:
                coverage_rev+=lead_provider.covrtab_rev[bin]

            coverage_sum+=coverage_fwd+coverage_rev
            bin_count+=1

            if bin%coverage_binsize_combine==0:
                block_index=int(bin/snf_block_size)*snf_block_size

                coverage_total_curr=math.ceil(coverage_sum/float(bin_count))
                if coverage_total_curr > 0:
                    if not block_index in self.blocks:
                        self.blocks[block_index]={svtype: [] for svtype in sv.TYPES}
                        self.blocks[block_index]["_COVERAGE"]={}

                    self.blocks[block_index]["_COVERAGE"][bin]=coverage_total_curr

                coverage_sum=0
                bin_count=0

    def serialize_block(self,block_id):
        return pickle.dumps(self.blocks[block_id])

    def unserialize_block(self,data):
        return pickle.loads(data)

    def write_and_index(self):
        offset=0
        for block_id in sorted(self.blocks):
            data=gzip.compress(self.serialize_block(block_id))
            self.handle.write(data)
            data_len=len(data)
            self.index[block_id]=(offset,data_len)
            offset+=data_len
            self.total_length+=data_len

    def read_header(self):
        try:
            header_text=self.handle.readline()
            self.header_length=len(header_text)
            self.header=json.loads(header_text.strip())
        except Exception as e:
            print(f"Error when reading SNF header from '{self.filename}': {e}. The file may not be a valid .snf file or could have been corrupted.")
            raise e
        self.index=self.header["index"]

    def read_blocks(self,contig,block_index):
        block_index=str(block_index)
        if not contig in self.index:
            return None

        if not block_index in self.index[contig]:
            return None

        blocks=[]
        for block_data_start,block_data_length in self.index[contig][block_index]:
            try:
                self.handle.seek(self.header_length+block_data_start)
                data=gzip.decompress(self.handle.read(block_data_length))
                blocks.append(self.unserialize_block(data))
            except Exception as e:
                print(f"Error when reading block '{contig}.{block_index}' from '{self.filename}': {e}. The file may not be a valid .snf file or could have been corrupted.")
                raise e
        return blocks

    def get_index(self):
        return self.index

    def get_total_length(self):
        return self.total_length

    def close(self):
        self.handle.close()



#!/usr/bin/env python3
#
# Sniffles2
# A fast structural variant caller for long-read sequencing data
#
# Created: 27.08.2021
# Author:  Moritz Smolka
# Contact: moritz.g.smolka@gmail.com
#

from dataclasses import dataclass

import collections

from sniffles import sv
from sniffles import util
from sniffles import consensus

import math

def annotate_sv(svcall,config):
    if config.phase:
        phase=phase_sv(svcall,config)
    else:
        phase=None

    genotype_sv(svcall,config,phase)

    if svcall.svtype=="INS" and not config.symbolic:
        merged_leads=[l for l in svcall.postprocess.cluster.leads if l.seq!=None]

        if len(merged_leads):
            best_lead=merged_leads[0]
            best_index=0
            best_diff=abs(len(best_lead.seq) - svcall.svlen) + abs(best_lead.ref_start - svcall.pos)*1.5
            for i,ld in enumerate(merged_leads):
                if i==0:
                    continue
                curr_diff=abs(len(ld.seq) - svcall.svlen) + abs(ld.ref_start - svcall.pos)*1.5
                if curr_diff < best_diff:
                    best_lead=ld
                    best_index=i
                    best_diff=curr_diff

            merged_leads.pop(best_index)
            #merged_leads_new=list()

            #for lead in merged_leads:
            #    curr_lendiff=abs(len(ld.seq) - len(best_lead.seq)) + abs(ld.ref_start - best_lead.ref_start)
            #    #if curr_lendiff < 25:
            #    merged_leads_new.append(lead)
            #merged_leads=merged_leads_new


            if len(merged_leads) >= config.consensus_min_reads and not config.no_consensus:
                klen=config.consensus_kmer_len
                skip=config.consensus_kmer_skip_base+int(len(best_lead.seq)*config.consensus_kmer_skip_seqlen_mult)
                skip_repetitive=skip

                svcall.alt=consensus.novel_from_reads(best_lead,merged_leads,klen=klen,skip=skip,skip_repetitive=skip_repetitive)
            else:
                svcall.alt=best_lead.seq

def add_request(svcall,field,pos,requests_for_coverage,config):
    bin=int(pos/config.coverage_binsize)*config.coverage_binsize
    if not bin in requests_for_coverage:
        requests_for_coverage[bin]=[]
    requests_for_coverage[bin].append((svcall,field))

def coverage(calls,lead_provider,config):
    requests_for_coverage=coverage_build_requests(calls,lead_provider,config)
    return coverage_fulfill(requests_for_coverage,calls,lead_provider,config)

def coverage_build_requests(calls,lead_provider,config):
    requests_for_coverage={}
    for svcall in calls:
        start=svcall.pos
        if svcall.svtype=="INS":
            end=start+1
        else:
            end=svcall.pos+abs(svcall.svlen)
        add_request(svcall,"coverage_start",start-config.coverage_binsize,requests_for_coverage,config)
        add_request(svcall,"coverage_center",int((start+end)/2),requests_for_coverage,config)
        add_request(svcall,"coverage_end",end+config.coverage_binsize,requests_for_coverage,config)
        add_request(svcall,"coverage_upstream",start-config.coverage_binsize*config.coverage_updown_bins,requests_for_coverage,config)
        add_request(svcall,"coverage_downstream",end+config.coverage_binsize*config.coverage_updown_bins,requests_for_coverage,config)
    return requests_for_coverage

def coverage_fulfill(requests_for_coverage,calls,lead_provider,config):
    if len(requests_for_coverage)==0:
        return -1,-1

    start_bin=lead_provider.covrtab_min_bin
    end_bin=int(lead_provider.end/config.coverage_binsize)*config.coverage_binsize
    coverage_fwd=0
    coverage_rev=0
    coverage_fwd_total=0
    coverage_rev_total=0
    n=0

    for bin in range(start_bin, end_bin+config.coverage_binsize,config.coverage_binsize):
        n+=1

        if bin in lead_provider.covrtab_fwd:
            coverage_fwd+=lead_provider.covrtab_fwd[bin]

        if bin in lead_provider.covrtab_rev:
            coverage_rev+=lead_provider.covrtab_rev[bin]

        if bin in requests_for_coverage:
            coverage_total_curr=coverage_fwd+coverage_rev
            for svcall, field in requests_for_coverage[bin]:
                setattr(svcall,field,coverage_total_curr)

        coverage_fwd_total+=coverage_fwd
        coverage_rev_total+=coverage_rev

    average_coverage_fwd=coverage_fwd_total/float(n) if n>0 else 0
    average_coverage_rev=coverage_rev_total/float(n) if n>0 else 0
    return average_coverage_fwd,average_coverage_rev

def qc_sv_support(svcall,coverage_global,config):
    if config.minsupport == "auto":
        if not qc_support_auto(svcall,coverage_global,config):
            svcall.filter="SUPPORT_MIN"
            return False
    else:
        if not qc_support_const(svcall,config):
            svcall.filter="SUPPORT_MIN"
            return False
    return True

def rescale_support(svcall,config):
    if svcall.svtype!="INS" or svcall.svlen < config.long_ins_length:
        return svcall.support
    else:
        base=svcall.support + svcall.get_info("SUPPORT_LONG")
        scale_factor=config.long_ins_rescale_mult*(float(svcall.svlen)/config.long_ins_length)
        return round(base*(config.long_ins_rescale_base+scale_factor))

def qc_support_auto(svcall,coverage_global,config):
    support=rescale_support(svcall,config)
    #if svcall.svtype=="INS":
    #    coverage_list=[svcall.coverage_center]
    #else:
    coverage_list=[svcall.coverage_upstream,svcall.coverage_downstream]
    coverage_list=[c for c in coverage_list if c!=None and c!=0]
    if len(coverage_list)==0:
        coverage_list=[svcall.coverage_start,svcall.coverage_center,svcall.coverage_end]
        coverage_list=[c for c in coverage_list if c!=None and c!=0]

    if len(coverage_list)==0:
        coverage_regional=coverage_global
    else:
        coverage_regional=round(sum(coverage_list)/len(coverage_list))
        if coverage_regional==0:
            coverage_regional=coverage_global
    coverage=(coverage_regional*config.minsupport_auto_regional_coverage_weight+coverage_global*(1.0-config.minsupport_auto_regional_coverage_weight))
    min_support=round(config.minsupport_auto_base+config.minsupport_auto_mult*coverage)
    return support >= min_support
    #return True

def qc_support_const(svcall,config):
    #svcall.set_info("MINSUPPORT",config.minsupport)
    return svcall.support >= config.minsupport

def qc_sv(svcall,config):
    if config.qc_stdev:
        stdev_pos=svcall.get_info("STDEV_POS")
        if stdev_pos > config.qc_stdev_abs_max:
            svcall.filter="STDEV_POS"
            return False
        if svcall.svtype!="BND" and stdev_pos / abs(svcall.svlen) > 2.0:
            svcall.filter="STDEV_POS"
            return False

        stdev_len = svcall.get_info("STDEV_LEN")
        if stdev_len != None:
            if svcall.svtype != "BND" and stdev_len / abs(svcall.svlen) > 1.0:
                svcall.filter="STDEV_LEN"
                return False
            if stdev_len > config.qc_stdev_abs_max:
                svcall.filter="STDEV_LEN"
                return False

    if abs(svcall.svlen) < config.minsvlen:
        svcall.filter="SVLEN_MIN"
        return False

    #if (svcall.coverage_upstream != None and svcall.coverage_upstream < config.qc_coverage) or (svcall.coverage_downstream != None and svcall.coverage_downstream < config.qc_coverage):
    if svcall.svtype != "DEL" and svcall.svtype != "INS" and (svcall.coverage_center != None and svcall.coverage_center < config.qc_coverage):
        svcall.filter="COV_MIN"
        return False

    if svcall.svtype == "DEL" and config.long_del_length != -1 and abs(svcall.svlen) >= config.long_del_length and not config.non_germline:
        if svcall.coverage_center != None and svcall.coverage_upstream != None and svcall.coverage_downstream != None and svcall.coverage_center > (svcall.coverage_upstream+svcall.coverage_downstream)/2.0 * config.long_del_coverage:
            svcall.filter="COV_CHANGE"
            return False
    elif svcall.svtype=="INS" and ( (svcall.coverage_upstream != None and svcall.coverage_upstream < config.qc_coverage) or (svcall.coverage_downstream != None and svcall.coverage_downstream < config.qc_coverage)):
        svcall.filter="COV_CHANGE"
        return False
    elif svcall.svtype == "DUP" and config.long_dup_length != -1 and abs(svcall.svlen) >= config.long_dup_length and not config.non_germline:
        if svcall.coverage_center != None and svcall.coverage_upstream != None and svcall.coverage_downstream != None and svcall.coverage_center < (svcall.coverage_upstream+svcall.coverage_downstream)/2.0 * config.long_dup_coverage:
            svcall.filter="COV_CHANGE"
            return False

    return True

def qc_sv_post_annotate(svcall,config):
    if (len(svcall.genotypes)==0 or (svcall.genotypes[0][0]!="." and svcall.genotypes[0][0]+svcall.genotypes[0][1]<2)) and (svcall.coverage_center != None and svcall.coverage_center < config.qc_coverage):
        svcall.filter="COV_MIN"
        return False

    return True

def binomial_coef(n,k):
    return math.factorial(n)/(math.factorial(k)*math.factorial(n-k))

def binomial_probability(k,n,p):
    try:
        #Binomial coef cancels out for likelihood ratios
        #return binomial_coef(n,k) * (p**k) * ((1.0-p)**(n-k))
        return (p**k) * ((1.0-p)**(n-k))
    except OverflowError:
        return 1.0

def likelihood_ratio(q1,q2):
    if q1/q2>0:
        try:
            return math.log(q1/q2,10)
        except ValueError:
            return 0
    else:
        return 0

def genotype_sv(svcall,config,phase):
    normalization_target=250
    hom_ref_p=config.genotype_error
    het_p=(1.0/config.genotype_ploidy) # - config.genotype_error
    hom_var_p=1.0 - config.genotype_error
    coverage=0

    #Count inline events only once per read, but split events as individual alignments, as in coverage calculation
    leads=svcall.postprocess.cluster.leads
    support=rescale_support(svcall,config)

    if svcall.svtype=="INS":
        coverage_list=[svcall.coverage_center]
    else:
        if svcall.svtype=="DUP":
            if False and svcall.coverage_start!=None and svcall.coverage_end!=None:
                if svcall.coverage_start>svcall.coverage_end:
                    coverage_list=[svcall.coverage_end]
                else:
                    coverage_list=[svcall.coverage_start]
            else:
                coverage_list=[svcall.coverage_start,svcall.coverage_end]
            coverage+=round(support*0.75)
        elif svcall.svtype=="INV":
            coverage_list=[svcall.coverage_upstream,svcall.coverage_downstream]
            coverage+=round(support*0.5)
        else:
            coverage_list=[svcall.coverage_start,svcall.coverage_center,svcall.coverage_end]

    coverage_list=[c for c in coverage_list if c!=None and c!=0]
    if len(coverage_list)==0:
        return
    coverage+=round(sum(coverage_list)/len(coverage_list))

    if support > coverage:
        coverage=support

    af=support / float(coverage)

    genotype_p=[((0,0),hom_ref_p),
                ((0,1),het_p),
                ((1,1),hom_var_p)]

    max_lead=max(support,coverage)
    if max_lead>normalization_target:
        norm=normalization_target/float(max_lead)
        normalized_support=round(support*norm)
        normalized_coverage=round(coverage*norm)
    else:
        normalized_support=support
        normalized_coverage=coverage

    genotype_likelihoods=[]
    for gt, p in genotype_p:
        q=binomial_probability(normalized_support,normalized_coverage,p)
        genotype_likelihoods.append((gt,q))
    genotype_likelihoods.sort(key=lambda k: k[1], reverse=True)

    sum_likelihoods=sum(q for gt,q in genotype_likelihoods)
    normalized_likelihoods=[ (gt,(q/sum_likelihoods)) for gt,q in genotype_likelihoods]

    gt1,q1=normalized_likelihoods[0]
    gt2,q2=normalized_likelihoods[1]
    qz=[q for gt,q in normalized_likelihoods if gt==(0,0)][0]
    genotype_z_score = min(60,int((-10) * likelihood_ratio(qz,q1)))
    genotype_quality = min(60,int((-10) * likelihood_ratio(q2,q1)))

    is_long_ins=(svcall.svtype=="INS" and svcall.svlen >= config.long_ins_length)
    if genotype_z_score < config.genotype_min_z_score and not config.non_germline and not is_long_ins:
        if svcall.filter=="PASS":
            svcall.filter="GT"

    if is_long_ins and gt1==(0,0):
        a,b=".","."
    else:
        a,b=gt1
    svcall.genotypes[0]=(a,b,genotype_quality,coverage-support,support,phase)
    svcall.set_info("AF",af)

def phase_sv(svcall,config):
    reads_phases={lead.read_id[0]: (lead.read_id[1],lead.read_id[2]) for lead in svcall.postprocess.cluster.leads}
    hp_list=util.most_common(hp for hp,ps in reads_phases.values())
    ps_list=util.most_common(ps for hp,ps in reads_phases.values())

    hp_support,hp=hp_list[0]
    ps_support,ps=ps_list[0]
    if hp==None:
        hp="NULL"
    if ps==None:
        ps="NULL"

    other_hp_support=sum(other_supp for other_supp, other_hp in hp_list if other_hp != hp and other_hp != "NULL")
    other_ps_support=sum(other_supp for other_supp, other_ps in ps_list if other_ps != ps and other_ps != "NULL")

    hp_filter="FAIL"
    if hp != "NULL" and hp_support > 0 and float(other_hp_support)/(hp_support+other_hp_support) < config.phase_conflict_threshold:
        hp_filter="PASS"

    ps_filter="FAIL"
    if ps != "NULL" and ps_support > 0 and float(other_ps_support)/(ps_support+other_ps_support) < config.phase_conflict_threshold:
        ps_filter="PASS"

    svcall.set_info("PHASE",f"{hp},{ps},{hp_support},{ps_support},{hp_filter},{ps_filter}")
    return (config.phase_identifiers.index(hp) if hp in config.phase_identifiers else None if hp_filter=="PASS" else None)



#!/usr/bin/env python3
#
# Sniffles2
# A fast structural variant caller for long-read sequencing data
#
# Created: 09.09.2021
# Author:  Moritz Smolka
# Contact: moritz.g.smolka@gmail.com
#

from dataclasses import dataclass
import collections

from sniffles import util

@dataclass
class Group:
    """
    A sequence to which other sequences are pseudoaligned (mappings) through exact kmer matches (anchors)
    Can be reduced to a consensus sequence using the flatten() method
    """
    anchors: dict
    mappings: dict
    consensus_min: int
    consensus_max: int
    start_len: int
    origin: object
    support: list

    def flatten(self,frameshift_origin=False,return_debug=False):
        override_origin_min=3
        override_origin_indel_ratio=0.6

        consensus=""
        origin_space_consensus=""
        ins_consensus=""
        del_consensus=""
        mappings_consensus=""
        skip_until=self.consensus_min-1
        for ci in range(self.consensus_min,self.consensus_max+1,1):
            if not ci in self.mappings:
                origin_space_consensus+="-"
                ins_consensus+="-"
                del_consensus+="-"
                mappings_consensus+="-"
                continue
            type_fragments={"M":[], "I":[], "D":[]}
            type_fragments_counts={"M":0, "I":0, "D":0}
            for maptype, maplen, map_last_i, map_i, map_j, map_lead in self.mappings[ci]:
                type_fragments[maptype].append((maplen,map_lead.seq[map_last_i:map_i]))
                type_fragments_counts[maptype]+=1

            type_consensus=sorted(type_fragments, key=lambda t: len(type_fragments[t]))[-1]
            assert(len(type_fragments[type_consensus])>0)

            #Favor match over I/D when ratio is below 2

            if type_consensus=="I" and type_fragments_counts["M"]>0 and type_fragments_counts["I"] / float(type_fragments_counts["M"]+type_fragments_counts["I"]) < override_origin_indel_ratio:
                type_consensus="N"

            if type_consensus=="D" and type_fragments_counts["M"]>0 and type_fragments_counts["D"] / float(type_fragments_counts["M"]+type_fragments_counts["D"]) < override_origin_indel_ratio:
                type_consensus="N"

            if type_consensus=="M":
                #print(type_fragments_seq)
                if len(type_fragments[type_consensus]) < 3:
                    #Favor original sequence when only two mappings present
                    maplen,seq=type_fragments[type_consensus][0]
                    common_base,count=seq[0],1
                else:
                    common_base,count=collections.Counter(seq[0] for maplen,seq in type_fragments[type_consensus]).most_common(1)[0]

                if ci > skip_until:
                    consensus+=common_base
                    origin_space_consensus+=common_base
                else:
                    origin_space_consensus+="-"
                ins_consensus+="-"
                del_consensus+="-"
                mappings_consensus+=str(min(9,count))

            elif type_consensus=="N":
                #Favor original sequence when only two mappings present
                maplen,seq=type_fragments["M"][0]
                common_base,count=seq[0],1

                if ci > skip_until:
                    consensus+=common_base
                    origin_space_consensus+=common_base
                else:
                    origin_space_consensus+="-"
                ins_consensus+="-"
                del_consensus+="-"
                mappings_consensus+=str(min(9,count))

            elif frameshift_origin and type_consensus=="I":
                common_maplen,count=collections.Counter(maplen for maplen,seq in type_fragments[type_consensus]).most_common(1)[0]
                common_seqlen,count=collections.Counter(len(seq) for maplen,seq in type_fragments[type_consensus] if maplen==common_maplen).most_common(1)[0]
                origin_space_consensus+="^"
                ins_consensus+=str(min(9,maplen))
                del_consensus+="-"
                mappings_consensus+=str(min(9,count))

                selected=[seq for maplen,seq in type_fragments[type_consensus] if common_maplen==maplen and len(seq)==common_seqlen]

                for i in range(common_seqlen):
                    common_ins_base,_=collections.Counter(seq[i] for seq in selected).most_common(1)[0]
                    consensus+=common_ins_base

                skip_until=ci+common_seqlen+common_maplen-1

            elif frameshift_origin and type_consensus=="D":
                del_consensus=del_consensus[:-1]+str(min(9,maplen))

        if not return_debug:
            return consensus
        else:
            return origin_space_consensus,ins_consensus,del_consensus,mappings_consensus,consensus

def make_group(lead,klen,skip,repetitive_kmers):
    anchors={}
    for i, kmer in iter_kmers(lead.seq,klen,skip):
        if kmer in repetitive_kmers:
            continue
        if not kmer in anchors:
            anchors[kmer]=[i]
        else:
            anchors[kmer].append(i)

    return Group(anchors=anchors,
                 mappings={i:[("M", 1, i, i+1, i, lead)] for i in range(len(lead.seq))},
                 consensus_min=0,
                 consensus_max=len(lead.seq)-1,
                 start_len=len(lead.seq),
                 origin=lead,
                 support=[lead])

def iter_kmers(seq,klen,skip):
    for i in range(0,len(seq)-klen,skip):
        yield (i,seq[i:i+klen])

def from_leads(leads, initial_lead, max_grp=3, klen=15, skip=1, skip_repetitive=1, extend=False):
    #TODO: Handle DELs in flattening
    #TODO: Allow mapping across putative INDELs at sequence ends, allowing correcting them
    #TODO: Dynamically adapt kmer len threshold so short+long origs can be assembled at the same time, same for repetitive_limit
    #TODO: Ensure that kmers before and after last mapping are added as (secondary) anchors (hints 4 indels)

    repetitive_limit=len(leads)+2

    #Generate repetitive kmer list
    kmer_counts={}
    #for lead in leads:
    for i, kmer in iter_kmers(initial_lead.seq,klen=klen,skip=skip_repetitive):
        if not kmer in kmer_counts:
            kmer_counts[kmer]=1
        else:
            kmer_counts[kmer]+=1
    repetitive_kmers=set(kmer for kmer,count in kmer_counts.items() if count>repetitive_limit)

    #Create initial group
    leads=sorted(leads,key=lambda ld: len(ld.seq))
    groups=[make_group(initial_lead,klen,skip,repetitive_kmers)]

    for lead in leads:
        if lead.seq==None:
            continue

        assigned=False

        #Try to place this lead's INS seq in the group
        for group in groups[:max_grp]:
            mapped=False
            last_i,last_j,last_was_repeat=None,None,False
            first_i,first_j=None,None

            #Attempt to anchor kmers from seq to current group
            for i, kmer in iter_kmers(lead.seq,klen=klen,skip=skip):
                if not kmer in group.anchors: #and len(set(group.anchors[kmer]))==1:
                    continue

                if len(set(group.anchors[kmer])) > 1:
                    if last_j==None:
                        #Do not allow the first kmer to have multiple potential positions
                        continue

                    #Kmer occurs at multiple points in group sequence. Attempt to choose anchor most suitable for linear mapping
                    j=group.anchors[kmer][0]
                    for anchor_j_curr in group.anchors[kmer][1:]:
                        if abs(anchor_j_curr-(last_j+1)) < abs(j-(last_j+1)):
                            j=anchor_j_curr
                    last_was_repeat=True
                else:
                    #Kmer can be unambiguously placed on group sequence
                    j=group.anchors[kmer][0]

                if first_i==None and first_j==None:
                    first_i=i
                    first_j=j

                if last_i != None and last_j != None:
                    #Determine mapping type by position of current and last kmer on the lead sequence and group sequence
                    dist_lead=i-last_i
                    dist_group=j-last_j

                    if not last_was_repeat and dist_lead > dist_group:
                        maptype="D"
                        maplen=dist_group-dist_lead
                        if not last_j in group.mappings:
                            group.mappings[last_j]=[]
                        group.mappings[last_j].append((maptype, maplen, last_i, i, j, lead))

                    elif not last_was_repeat and dist_lead < dist_group:
                        maptype="I"
                        maplen=dist_group-dist_lead
                        if not last_j in group.mappings:
                            group.mappings[last_j]=[]
                        group.mappings[last_j].append((maptype, maplen, last_i, i, j, lead))

                    else:
                        maptype="M"
                        maplen=last_i-i
                        act_j=last_j
                        for act_i in range(last_i,i):
                            if not act_j in group.mappings:
                                group.mappings[act_j]=[]
                            group.mappings[act_j].append((maptype, maplen, act_i, act_i+1, last_j, lead))
                            act_j+=1

                mapped=True
                last_i=i
                last_j=j
                last_was_repeat=False

            if mapped:
                group.support.append(lead)
                assigned=True
                max_i=len(lead.seq)

                #Extend consensus group leftwards
                if extend and first_i - (first_j-group.consensus_min) > 0: # and first_j < 1.5*klen:
                    for i in range(first_i,-1,-1):
                        curr_j=first_j-(first_i-i)
                        if curr_j <= group.consensus_min: #+klen/1:
                            assert(curr_j==group.consensus_min)
                            kmer=lead.seq[i:i+klen]
                            if not kmer in repetitive_kmers:
                                group.anchors[kmer]=[curr_j]
                            group.consensus_min-=1
                            if not curr_j in group.mappings:
                                group.mappings[curr_j]=[]
                            group.mappings[curr_j].append(("M", 1, i, i+1, first_j, lead))

                #Extend consensus group rightwards
                if extend and last_j+(max_i-last_i) > group.consensus_max:
                    for i in range(last_i,max_i,1):
                        curr_j=last_j+(i-last_i)
                        if curr_j+klen >= group.consensus_max: #-klen/1:
                            assert(curr_j+klen==group.consensus_max)
                            kmer=lead.seq[i:i+klen]
                            if not kmer in repetitive_kmers:
                                group.anchors[kmer]=[curr_j]
                            group.consensus_max+=1
                            if not curr_j in group.mappings:
                                group.mappings[curr_j]=[]
                            group.mappings[curr_j].append(("M", 1, i, i+1, last_j, lead))
                break

        if not assigned:
            groups.append(make_group(lead,klen,skip,repetitive_kmers))
            #groups.sort(key=lambda g: len(g.mappings), reverse=True)

    return groups


def novel_from_reads(best_lead,other_leads,klen,skip,skip_repetitive,debug=False):
    consensus_min=2
    maxshift=klen
    minspan=0.2
    minalns=0.25
    minident=0.5
    minident_abs=5
    minbestdiff=3

    alignments=[]
    anchors={}
    taboo=set()
    for i, kmer in iter_kmers(best_lead.seq,klen=klen,skip=skip_repetitive):
        if kmer in taboo:
            continue
        if kmer in anchors:
            del anchors[kmer]
            taboo.add(kmer)
            continue
        anchors[kmer]=i

    for leadi,lead in enumerate(other_leads):
        last_i=None
        last_j=None
        conseq=""
        span=0
        for j, kmer in iter_kmers(lead.seq,klen=klen,skip=skip):
            if not kmer in anchors:
                continue
            i=anchors[kmer]
            if abs(i-j) > maxshift:
                continue
            if last_i != None and i <= last_i:
                continue

            if last_i == None:
                if j>0:
                    conseq="-"*i
            else:
                fwd_i=i-last_i
                fwd_j=j-last_j
                if len(conseq)+fwd_j > len(best_lead.seq):
                    fwd_j=len(best_lead.seq)-len(conseq)

                if fwd_i == fwd_j and fwd_j > 0:
                    span+=(j-last_j)
                    m=0
                    for l in range(1,(j-last_j)+1):
                        if lead.seq[last_j+l] == best_lead.seq[last_i+l]:
                            m+=1
                    ident=m/float((j-last_j))
                    if ident >= minident:
                        conseq+=lead.seq[last_j:j][:fwd_j]
                    else:
                        conseq+="-"*(fwd_j)
                else:
                    conseq+="-"*(fwd_j)
            last_i=i
            last_j=j

        if len(conseq) < len(best_lead.seq):
            conseq+="-"*(len(best_lead.seq)-len(conseq))

        conseq_new=[]
        h=0
        while h < len(best_lead.seq):
            if conseq[h]=="-":
                conseq_new.append("-")
                h+=1
            else:
                buffer=[]
                ident=0
                while h < len(best_lead.seq) and conseq[h]!="-":
                    ident+=(best_lead.seq[h]==conseq[h])
                    buffer.append(conseq[h])
                    h+=1
                if ident/float(len(buffer)) > minident and ident>minident_abs:
                    conseq_new.append("".join(buffer))
                else:
                    conseq_new.append("-"*len(buffer))
        conseq="".join(conseq_new)

        if span/float(len(best_lead.seq)) > minspan:
            alignments.append(conseq)

    maxal=1
    for i in range(len(best_lead.seq)):
        maxal=max(maxal,len([best_lead.seq[i]]+[a[i] for a in alignments if not a[i] in "^_"]))
    maxal=float(maxal)

    flattened=""
    for i in range(len(best_lead.seq)):
        al=[a[i] for a in alignments if not a[i]=="-"]
        if len(al) < consensus_min or len(al)/maxal < minalns:
            flattened+=best_lead.seq[i]
        else:
            top=util.most_common([best_lead.seq[i]]+al)
            if len(top)>1 and top[0][0]-top[1][0] >= minbestdiff:
                flattened+=top[0][1]
            else:
                flattened+=best_lead.seq[i]

            #if top[0][0]/float(len(al)+1) < 0.75:
            #    flattened+=best_lead.seq[i]
            #else:
            #    flattened+=top[0][1]

    #print("FLT",flattened)
    #if debug:
    #    print("B",best_lead.seq)
    #    for o in other_leads:
    #        print("O",o.seq)
    #    print("F",flattened)
    #    print("=====")
    return flattened
